import importlib


class DepotManager(object):
    """Takes care of managing the whole Depot environment for the application.

    DepotManager tracks the created depots, the current default depot,
    and the WSGI middleware in charge of serving files for local depots.

    While this is used to create the default depot used by the application it can
    also create additional depots using the :meth:`new` method.

    In case you need to migrate your application to a different storage while
    keeping compatibility with previously stored file simply change the default depot
    through :meth:`set_default` all previously stored file will continue to work
    on the old depot while new files will be uploaded to the new default one.

    """
    _default_depot = None
    _depots = {}
    _middleware = None
    _aliases = {}

    @classmethod
    def set_default(cls, name):
        """Replaces the current application default depot"""
        if name not in cls._depots:
            raise RuntimeError('%s depot has not been configured' % (name,))
        cls._default_depot = name

    @classmethod
    def get_default(cls):
        """Retrieves the current application default depot"""
        if cls._default_depot is None:
            raise RuntimeError('No depots have been configured!')
        return cls._default_depot

    @classmethod
    def set_middleware(cls, mw):
        if cls._middleware is not None:
            raise RuntimeError('There is already a WSGI middleware registered')
        cls._middleware = mw

    @classmethod
    def get_middleware(cls):
        if cls._middleware is None:
            raise RuntimeError('No WSGI middleware currently registered')
        return cls._middleware

    @classmethod
    def get(cls, name=None):
        """Gets the application wide depot instance.

        Might return ``None`` if :meth:`configure` has not been
        called yet.

        """
        if name is None:
            name = cls._default_depot

        name = cls.resolve_alias(name)  # resolve alias
        return cls._depots.get(name)

    @classmethod
    def get_file(cls, path):
        """Retrieves a file by storage name and fileid in the form of a path

        Path is expected to be ``storage_name/fileid``.
        """
        depot_name, file_id = path.split('/', 1)
        depot = cls.get(depot_name)
        return depot.get(file_id)

    @classmethod
    def url_for(cls, path):
        """Given path of a file uploaded on depot returns the url that serves it

        Path is expected to be ``storage_name/fileid``.
        """
        mw = cls.get_middleware()
        return mw.url_for(path)

    @classmethod
    def configure(cls, name, config, prefix='depot.'):
        """Configures an application depot.

        This configures the application wide depot from a settings dictionary.
        The settings dictionary is usually loaded from an application configuration
        file where all the depot options are specified with a given ``prefix``.

        The default ``prefix`` is *depot.*, the minimum required setting
        is ``depot.backend`` which specified the required backend for files storage.
        Additional options depend on the choosen backend.

        """
        if name in cls._depots:
            raise RuntimeError('Depot %s has already been configured' % (name,))

        if cls._default_depot is None:
            cls._default_depot = name

        cls._depots[name] = cls.from_config(config, prefix)
        return cls._depots[name]

    @classmethod
    def alias(cls, alias, name):
        if name not in cls._depots:
            raise ValueError('You can only alias an existing storage, %s not found' % (name, ))

        if alias in cls._depots:
            raise ValueError('Cannot use an existing storage name as an alias, will break existing files.')

        cls._aliases[alias] = name

    @classmethod
    def resolve_alias(cls, name):
        while name and name not in cls._depots:
            name = cls._aliases.get(name)
        return name

    @classmethod
    def make_middleware(cls, app, **options):
        """Creates the application WSGI middleware in charge of serving local files.

        A Depot middleware is required if your application wants to serve files from
        storages that don't directly provide and HTTP interface like
        :class:`depot.io.local.LocalFileStorage` and :class:`depot.io.gridfs.GridFSStorage`

        """
        from depot.middleware import DepotMiddleware
        mw = DepotMiddleware(app, **options)
        cls.set_middleware(mw)
        return mw

    @classmethod
    def _new(cls, backend, **options):
        module, classname = backend.rsplit('.', 1)
        backend = importlib.import_module(module)
        class_ = getattr(backend, classname)
        return class_(**options)

    @classmethod
    def from_config(cls, config, prefix='depot.'):
        """Creates a new depot from a settings dictionary.

        Behaves like the :meth:`configure` method but instead of configuring the application
        depot it creates a new one each time.
        """
        config = config or {}

        # Get preferred storage backend
        backend = config.get(prefix + 'backend', 'depot.io.local.LocalFileStorage')

        # Get all options
        prefixlen = len(prefix)
        options = dict((k[prefixlen:], config[k]) for k in config.keys() if k.startswith(prefix))

        # Backend is already passed as a positional argument
        options.pop('backend', None)
        return cls._new(backend, **options)

    @classmethod
    def _clear(cls):
        """This is only for testing pourposes, resets the DepotManager status

        This is to simplify writing test fixtures, resets the DepotManager global
        status and removes the informations related to the current configured depots
        and middleware.
        """
        cls._default_depot = None
        cls._depots = {}
        cls._middleware = None
        cls._aliases = {}

get_depot = DepotManager.get
get_file = DepotManager.get_file
configure = DepotManager.configure
set_default = DepotManager.set_default



"""
Provides FileStorage implementation for MongoDB GridFS.

This is useful for storing files inside a mongodb database.

"""
from __future__ import absolute_import

from datetime import datetime
from pymongo import MongoClient
import gridfs
from bson import ObjectId

from .interfaces import FileStorage, StoredFile
from . import utils


class GridFSStoredFile(StoredFile):
    def __init__(self, file_id, gridout):
        _check_file_id(file_id)
        self._gridout = gridout
        self._closed = False

        metadata_info = {'filename': gridout.filename,
                         'content_type': gridout.content_type,
                         'content_length': gridout.length,
                         'last_modified': None}

        try:
            last_modified = gridout.last_modified
            if last_modified:
                metadata_info['last_modified'] = datetime.strptime(last_modified,
                                                                   '%Y-%m-%d %H:%M:%S')
        except:
            pass

        super(GridFSStoredFile, self).__init__(file_id=file_id, **metadata_info)

    def read(self, n=-1):
        if self._closed:
            raise ValueError("cannot read from a closed file")

        return self._gridout.read(n)

    def close(self):
        self._closed = True
        self._gridout.close()

    @property
    def closed(self):
        return self._closed


class GridFSStorage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that stores files on MongoDB.

    All the files are stored using GridFS to the database pointed by the ``mongouri`` parameter into
    the collection named ``collection``.

    """
    def __init__(self, mongouri, collection='filedepot'):
        self._cli = MongoClient(mongouri)
        self._db = self._cli.get_default_database()
        self._collection = collection
        self._gridfs = gridfs.GridFS(self._db, collection=collection)

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)

        try:
            gridout = self._gridfs.get(_check_file_id(fileid))
        except gridfs.errors.NoFile:
            raise IOError('File %s not existing' % fileid)

        return GridFSStoredFile(fileid, gridout)

    def create(self, content, filename=None, content_type=None):
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        new_file_id = self._gridfs.put(content,
                                       filename=filename,
                                       content_type=content_type,
                                       last_modified=utils.timestamp())
        return str(new_file_id)

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        fileid = _check_file_id(fileid)

        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(fileid):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(fileid))

        self._gridfs.delete(fileid)
        new_file_id = self._gridfs.put(content, _id=fileid,
                                       filename=filename,
                                       content_type=content_type,
                                       last_modified=utils.timestamp())
        return str(new_file_id)

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        fileid = _check_file_id(fileid)
        self._gridfs.delete(fileid)

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        fileid = _check_file_id(fileid)
        return self._gridfs.exists(fileid)

    def list(self):
        list_ids = []
        for filename in self._gridfs.list():
            list_ids.append(str(self._gridfs.find_one({"filename": filename})._id))
        return list_ids


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        return ObjectId(file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)



import unittest
from depot.io.gridfs import GridFSStorage

FILE_CONTENT = b'HELLO WORLD'


class TestGridFSFileStorage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            import pymongo.errors
            cls.fs = GridFSStorage('mongodb://localhost/gridfs_example?serverSelectionTimeoutMS=1', 'testfs')
            cls.fs._gridfs.exists("")  # Any operation to test that mongodb is up.
        except pymongo.errors.ConnectionFailure:
            cls.fs = None
    
    def setUp(self):
        if self.fs is None:
            self.skipTest('MongoDB not running')

    def tearDown(self):
        self.fs._db.drop_collection('testfs')

    def test_fileoutside_depot(self):
        fileid = self.fs._gridfs.put(FILE_CONTENT)

        f = self.fs.get(str(fileid))
        assert f.read() == FILE_CONTENT



from setuptools import setup, find_packages
import os, sys

version = '0.11.0'

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
except IOError:
    README = ''

TEST_DEPENDENCIES = ['mock',
                     'requests',
                     'ming',
                     'boto3',
                     'google-cloud-storage',
                     'flaky']

py_version = sys.version_info[:2]
if py_version != (3, 2):
    TEST_DEPENDENCIES += ['boto', 'coverage']
else:
    TEST_DEPENDENCIES += ['coverage < 4.0']

INSTALL_DEPENDENCIES = []
if py_version >= (3, 0):
    INSTALL_DEPENDENCIES += ["anyascii"]

if py_version == (2, 6):
    INSTALL_DEPENDENCIES += ['importlib']
    TEST_DEPENDENCIES += ['ordereddict', 'pillow < 4.0.0', 'WebTest < 2.0.24', 'sqlalchemy < 1.2']
else:
    TEST_DEPENDENCIES += ['pillow', 'WebTest', 'sqlalchemy']



setup(name='filedepot',
      version=version,
      description="Toolkit for storing files and attachments in web applications",
      long_description=README,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Environment :: Web Environment",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2"
        ],
      keywords='storage files s3 gridfs mongodb aws sqlalchemy wsgi',
      author='Alessandro Molina',
      author_email='amol@turbogears.org',
      url='https://github.com/amol-/depot',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'tests']),
      include_package_data=True,
      install_requires=INSTALL_DEPENDENCIES,
      tests_require=TEST_DEPENDENCIES,
      extras_require={
         'testing': TEST_DEPENDENCIES,
      },
      zip_safe=False,
)



import os
import uuid
import time
import unittest
import mock
import requests
from flaky import flaky
from unittest import SkipTest

from depot._compat import PY2, unicode_text


S3Storage = None
FILE_CONTENT = b'HELLO WORLD'


@flaky
class TestS3FileStorage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Travis runs multiple tests concurrently on fake machines that might
        # collide on pid and hostid, so use an uuid1 which should be fairly random
        # thanks to clock_seq
        cls.run_id = '%s-%s' % (uuid.uuid1().hex, os.getpid())

        try:
            global S3Storage
            from depot.io.awss3 import S3Storage
        except ImportError:
            raise SkipTest('Boto not installed')

        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
            raise SkipTest('Amazon S3 credentials not available')

        cls.default_bucket_name = 'filedepot-%s' % (access_key_id.lower(), )
        cls.cred = (access_key_id, secret_access_key)

        bucket_name = 'filedepot-testfs-%s' % cls.run_id
        cls.fs = S3Storage(access_key_id, secret_access_key, bucket_name)
        while not cls.fs._conn.lookup(bucket_name):
            # Wait for bucket to exist, to avoid flaky tests...
            time.sleep(0.5)

    @classmethod
    def tearDownClass(cls):
        if not cls.fs._conn.lookup(cls.fs._bucket_driver.bucket.name):
            return
        
        keys = [key.name for key in cls.fs._bucket_driver.bucket]
        if keys:
            cls.fs._bucket_driver.bucket.delete_keys(keys)

        try:
            cls.fs._conn.delete_bucket(cls.fs._bucket_driver.bucket.name)
            while cls.fs._conn.lookup(cls.fs._bucket_driver.bucket.name):
                # Wait for bucket to be deleted, to avoid flaky tests...
                time.sleep(0.5)
        except:
            pass

    def test_fileoutside_depot(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.set_contents_from_string(FILE_CONTENT)

        f = self.fs.get(fid)
        assert f.read() == FILE_CONTENT

    def test_invalid_modified(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.set_metadata('x-depot-modified', 'INVALID')
        key.set_contents_from_string(FILE_CONTENT)

        f = self.fs.get(fid)
        assert f.last_modified is None, f.last_modified

    def test_creates_bucket_when_missing(self):
        with mock.patch('boto.s3.connection.S3Connection.lookup', return_value=None):
            with mock.patch('boto.s3.connection.S3Connection.lookup',
                            return_value='YES') as mock_create:
                fs = S3Storage(*self.cred)
                mock_create.assert_called_with(self.default_bucket_name)

    def test_default_bucket_name(self):
        with mock.patch('boto.s3.connection.S3Connection.lookup', return_value='YES'):
            fs = S3Storage(*self.cred)
            assert fs._bucket_driver.bucket == 'YES'

    def test_public_url(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.set_contents_from_string(FILE_CONTENT)

        f = self.fs.get(fid)
        assert '.s3.amazonaws.com' in f.public_url, f.public_url
        assert f.public_url.endswith('/%s' % fid), f.public_url

    def test_content_disposition(self):
        file_id = self.fs.create(b'content', unicode_text('test.txt'), 'text/plain')
        test_file = self.fs.get(file_id)
        response = requests.get(test_file.public_url)
        assert response.headers['Content-Disposition'] == "inline;filename=\"test.txt\";filename*=utf-8''test.txt"



"""
Provides FileStorage implementation for local filesystem.

This is usefull for storing files inside a local path.

"""
import os
import uuid
import shutil
import json
from datetime import datetime

from .interfaces import FileStorage, StoredFile
from . import utils
from .._compat import unicode_text


class LocalStoredFile(StoredFile):
    def __init__(self, file_id, local_path):
        _check_file_id(file_id)

        self._metadata_path = _metadata_path(local_path)
        self._file_path = _file_path(local_path)
        self._file = None

        try:
            metadata = open(self._metadata_path, 'r')
        except:
            raise IOError('File %s not existing' % file_id)

        metadata_info = {'filename': 'unnamed',
                         'content_type': 'application/octet-stream',
                         'last_modified': None}
        with metadata:
            try:
                metadata_content = metadata.read()
                metadata_info.update(json.loads(metadata_content))

                last_modified = metadata_info['last_modified']
                if last_modified:
                    metadata_info['last_modified'] = datetime.strptime(last_modified,
                                                                       '%Y-%m-%d %H:%M:%S')
            except Exception:
                raise ValueError('Invalid file metadata for %s' % file_id)

        super(LocalStoredFile, self).__init__(file_id=file_id, **metadata_info)

    def read(self, n=-1):
        if self._file is None:
            self._file = open(self._file_path, 'rb')
        return self._file.read(n)

    def close(self):
        if self._file is None:
            # This is to guarantee that closing a file
            # before even reading it behaves correctly
            self._file = open(self._file_path, 'rb')
        self._file.close()

    @property
    def closed(self):
        if self._file is None:
            return False
        return self._file.closed


class LocalFileStorage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that stores files locally.

    All the files are stored inside a directory specified by the ``storage_path`` parameter.

    """
    def __init__(self, storage_path):
        self.storage_path = storage_path

    def __local_path(self, fileid):
        return os.path.join(self.storage_path, fileid)

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)
        local_file_path = self.__local_path(fileid)
        return LocalStoredFile(fileid, local_file_path)

    def __save_file(self, file_id, content, filename, content_type=None):
        local_file_path = self.__local_path(file_id)
        os.makedirs(local_file_path)
        saved_file_path = _file_path(local_file_path)

        if hasattr(content, 'read'):
            with open(saved_file_path, 'wb') as fileobj:
                shutil.copyfileobj(content, fileobj)
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')

            with open(saved_file_path, 'wb') as fileobj:
                fileobj.write(content)
                fileobj.flush()

        metadata = {'filename': filename,
                    'content_type': content_type,
                    'content_length': os.path.getsize(saved_file_path),
                    'last_modified': utils.timestamp()}

        with open(_metadata_path(local_file_path), 'w') as metadatafile:
            metadatafile.write(json.dumps(metadata))

    def create(self, content, filename=None, content_type=None):
        new_file_id = str(uuid.uuid1())
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        self.__save_file(new_file_id, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(fileid):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(fileid))

        self.delete(fileid)
        self.__save_file(fileid, content, filename, content_type)
        return fileid

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        local_file_path = self.__local_path(fileid)
        try:
            shutil.rmtree(local_file_path)
        except:
            pass

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        local_file_path = self.__local_path(fileid)
        return os.path.exists(local_file_path)

    def list(self):
        return [os.path.basename(fileid) for fileid in os.listdir(self.storage_path)]


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)


def _metadata_path(local_path):
    return os.path.join(local_path, 'metadata.json')


def _file_path(local_path):
    return os.path.join(local_path, 'file')



import io
import os
import uuid
import json
import requests
from flaky import flaky
import unittest
from google.cloud.exceptions import NotFound

from depot._compat import PY2, unicode_text
from depot.io.gcs import GCSStorage

FILE_CONTENT = b'HELLO WORLD'


@flaky
class TestGCSStorage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.run_id = '%s-%s' % (uuid.uuid1().hex, os.getpid())
        cls._prefix = ''

        try:
            global GCSStorage
            from depot.io.gcs import GCSStorage
        except ImportError:
            raise unittest.SkipTest('Google Cloud Storage not installed')
        
        cls._bucket = 'filedepot-testfs-%s' % cls.run_id
        env = os.environ
        google_credentials = env.get('GOOGLE_SERVICE_CREDENTIALS')
        if not google_credentials:
            raise unittest.SkipTest('GOOGLE_SERVICE_CREDENTIALS environment variable not set')
        google_credentials = json.loads(google_credentials)
        cls.fs = GCSStorage(project_id=google_credentials["project_id"], 
                            credentials=google_credentials, 
                            bucket=cls._bucket, 
                            prefix=cls._prefix)

    @classmethod
    def tearDownClass(cls):
        for blob in cls.fs.bucket.list_blobs():
            blob.delete()
        
        try:
            cls.fs.bucket.delete()
        except NotFound:
            pass

    def test_fileoutside_depot(self):
        fid = str(uuid.uuid1())
        blob = self.fs.bucket.blob(self._prefix+fid)
        blob.upload_from_string(FILE_CONTENT)

        f = self.fs.get(fid)
        assert f.read() == FILE_CONTENT

    def test_public_url(self):
        fid = str(uuid.uuid1())
        blob = self.fs.bucket.blob(self._prefix+fid)
        blob.upload_from_string(FILE_CONTENT)

        f = self.fs.get(fid)
        assert f.public_url.endswith('/%s' % self._prefix+fid), f.public_url

    def test_content_disposition(self):
        file_id = self.fs.create(io.BytesIO(b'content'), unicode_text('test.txt'), 'text/plain')
        test_file = self.fs.get(file_id)
        response = requests.get(test_file.public_url)
        assert response.headers['Content-Disposition'] == "inline;filename=\"test.txt\";filename*=utf-8''test.txt"

    def test_storage_non_ascii_filenames(self):
        filename = u'ملف.pdf'
        new_file_id = self.fs.create(
            io.BytesIO(FILE_CONTENT),
            filename=filename,
            content_type='application/pdf'
        )

        assert new_file_id is not None
    



from depot.manager import DepotManager
from .interfaces import DepotFileInfo
import json


class UploadedFile(DepotFileInfo):
    """Simple :class:`depot.fields.interfaces.DepotFileInfo` implementation that stores files.

    Takes a file as content and uploads it to the depot while saving around
    most file information. Pay attention that if the file gets replaced
    through depot manually the ``UploadedFile`` will continue to have the old data.

    Also provides support for encoding/decoding using JSON for storage inside
    databases as a plain string.

    Default attributes provided for all ``UploadedFile`` include:
        - filename     - This is the name of the uploaded file
        - file_id      - This is the ID of the uploaded file
        - path         - This is a depot_name/file_id path which can
                         be used with :meth:`DepotManager.get_file` to retrieve the file
        - content_type - This is the content type of the uploaded file
        - uploaded_at  - This is the upload date in YYYY-MM-DD HH:MM:SS format
        - url          - Public url of the uploaded file
        - file         - The :class:`depot.io.interfaces.StoredFile` instance of the stored file
    """
    def process_content(self, content, filename=None, content_type=None):
        """Standard implementation of :meth:`.DepotFileInfo.process_content`

        This is the standard depot implementation of files upload, it will
        store the file on the default depot and will provide the standard
        attributes.

        Subclasses will need to call this method to ensure the standard
        set of attributes is provided.
        """

        file_path, file_id = self.store_content(content, filename, content_type)
        self['file_id'] = file_id
        self['path'] = file_path

        saved_file = self.file
        self['filename'] = saved_file.filename
        self['content_type'] = saved_file.content_type
        self['uploaded_at'] = saved_file.last_modified.strftime('%Y-%m-%d %H:%M:%S')
        self['_public_url'] = saved_file.public_url

    def store_content(self, content, filename=None, content_type=None):
        file_id = self.depot.create(content, filename, content_type)
        file_path = '%s/%s' % (self.depot_name, file_id)
        self.files.append(file_path)
        return file_path, file_id

    def encode(self):
        return json.dumps(self)

    @classmethod
    def decode(cls, data):
        return cls(json.loads(data))

    @property
    def url(self):
        public_url = self['_public_url']
        if public_url:
            return public_url
        return DepotManager.get_middleware().url_for(self['path'])

    @property
    def depot(self):
        return DepotManager.get(self.depot_name)

    @property
    def file(self):
        return self.depot.get(self.file_id)


from datetime import datetime
from email.utils import parsedate_tz, mktime_tz
from time import gmtime, time
from .manager import DepotManager
from ._compat import wsgi_string
from .utils import make_content_disposition

_BLOCK_SIZE = 4096 * 64 # 256K


class _FileIter(object):
    def __init__(self, file, block_size):
        self.file = file
        self.block_size = block_size

    def __iter__(self):
        return self

    def next(self):
        val = self.file.read(self.block_size)
        if not val:
            raise StopIteration
        return val

    __next__ = next # py3

    def close(self):
        self.file.close()


class FileServeApp(object):
    """
    Serves a static filelike object.
    """
    def __init__(self, storedfile, cache_max_age, replace_wsgi_filewrapper=False):
        self.file = storedfile

        self.filename = self.file.filename
        self.last_modified = self.file.last_modified
        self.content_length = self.file.content_length
        self.content_type = self.file.content_type
        self.cache_expires = cache_max_age
        self.replace_wsgi_filewrapper = replace_wsgi_filewrapper

    def generate_etag(self):
        return '"%s-%s"' % (self.last_modified, self.content_length)

    def parse_date(self, value):
        try:
            return datetime.utcfromtimestamp(mktime_tz(parsedate_tz(value)))
        except (TypeError, OverflowError):
            raise RuntimeError("Received an ill-formed timestamp")

    @classmethod
    def make_date(cls, d):
        if isinstance(d, datetime):
            d = d.utctimetuple()
        else:
            d = gmtime(d)

        return '%s, %02d%s%s%s%s %02d:%02d:%02d GMT' % (
            ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')[d.tm_wday],
            d.tm_mday, ' ',
            ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
             'Oct', 'Nov', 'Dec')[d.tm_mon - 1],
            ' ', str(d.tm_year), d.tm_hour, d.tm_min, d.tm_sec)

    def has_been_modified(self, environ, etag, last_modified):
        unmodified = False

        modified_since = environ.get('HTTP_IF_MODIFIED_SINCE')
        if modified_since:
            modified_since = self.parse_date(modified_since)
            if last_modified and last_modified <= modified_since:
                unmodified = True

        if_none_match = environ.get('HTTP_IF_NONE_MATCH')
        if if_none_match and etag == if_none_match:
            unmodified = True

        return not unmodified

    def __call__(self, environ, start_response):
        headers = []
        timeout = self.cache_expires
        etag = self.generate_etag()
        headers += [('ETag', '%s' % etag),
            ('Cache-Control', 'max-age=%d, public' % timeout)]

        try:
            has_been_modified = self.has_been_modified(environ, etag, self.last_modified)
        except RuntimeError:
            start_response('400 Bad Request', [('Content-Type', 'text/html')])
            return ['''\
<html>
 <head>
  <title>400 Bad Request</title>
 </head>
 <body>
  <h1>400 Bad Request</h1>
  ETag or If-Modified-Since headers were malformed in request
 </body>
</html>'''.encode('ascii')]

        if not has_been_modified:
            self.file.close()
            start_response('304 Not Modified', headers)
            return []

        headers.extend((
            ('Expires', self.make_date(time() + timeout)),
            ('Content-Type', str(self.content_type)),
            ('Content-Length', str(self.content_length)),
            ('Last-Modified', self.make_date(self.last_modified)),
            ('Content-Disposition', make_content_disposition('inline', self.filename))
        ))
        start_response('200 OK', headers)

        if self.replace_wsgi_filewrapper is True:
            environ['wsgi.file_wrapper'] = _FileIter

        return environ.get('wsgi.file_wrapper', _FileIter)(self.file, _BLOCK_SIZE)


class DepotMiddleware(object):
    """WSGI Middleware in charge of serving Depot files.

    Usually created using :meth:`depot.manager.DepotManager.make_middleware`,
    it's a WSGI middleware that serves files stored inside depots that do not
    provide a public HTTP url. For depot that provide a public url the
    request is redirected to the public url.

    In case you have issues serving files with your WSGI server your can try
    to set ``replace_wsgi_filewrapper=True`` which forces DEPOT to use its own
    internal FileWrapper instead of the one provided by your WSGI server.

    """
    def __init__(self, app, mountpoint='/depot', cache_max_age=3600*24*7,
                 replace_wsgi_filewrapper=False):
        if not mountpoint.startswith('/'):
            raise ValueError('DepotMiddleware mountpoint must be an absolute path')

        self.app = app
        self.mountpoint = mountpoint
        self.cache_max_age = cache_max_age
        self.replace_wsgi_filewrapper = replace_wsgi_filewrapper

    def url_for(self, path):
        return wsgi_string('/'.join((self.mountpoint, path)))

    def _404_response(self, start_response):
        start_response('404 Not Found', [('Content-Type', 'text/html')])
        return ['''\
        <html>
         <head>
          <title>404 Not Found</title>
         </head>
         <body>
          <h1>404 Not Found</h1>
          File Not Found
         </body>
        </html>'''.encode('ascii')]

    def _301_response(self, start_response, location):
        # Should also set Cache-Control to keep around the 301
        start_response('301 Moved Permanently', [('Content-Type', 'text/html'),
                                                 ('Location', wsgi_string(location))])
        return [('''\
        <html>
         <head>
          <title>301 Moved Permanently</title>
         </head>
         <body>
          <h1>301 Moved Permanently</h1>
          File you are looking for is available at <a href="%s">%s</a>
         </body>
        </html>''' % (location, location)).encode('ascii')]

    def __call__(self, environ, start_response):
        req_method = environ['REQUEST_METHOD']
        full_path = environ['PATH_INFO']

        mtpointlen = len(self.mountpoint)
        if not (req_method in ('GET', 'HEAD') and full_path.startswith(self.mountpoint) and 
                full_path[mtpointlen:mtpointlen+1] in ('', '/')):
            return self.app(environ, start_response)

        
        path = full_path.rsplit('/', 2)
        if len(path) and not path[0]:
            path = path[1:]


        if len(path) < 3:
            return self._404_response(start_response)

        __, depot, fileid = path[:3]
        depot = DepotManager.get(depot)
        if not depot:
            return self._404_response(start_response)

        try:
            f = depot.get(fileid)
        except (IOError, ValueError):
            return self._404_response(start_response)

        public_url = f.public_url
        if public_url is not None:
            return self._301_response(start_response, public_url)

        fileapp = FileServeApp(f, self.cache_max_age, self.replace_wsgi_filewrapper)
        return fileapp(environ, start_response)
